import datetime
import decimal

import dash_devextreme as ddx
import dash_core_components as dcc

from django import forms
from django.db import models
from django.utils import timezone


__all__ = (
    'ddx_table_columns',
    'ddx_widgets',
    'display_for_field',
    'display_for_value',
)


def column_bool(data_field, caption, **kwargs):
    return dict(dataField=data_field, caption=caption, dataType='boolean', **kwargs)


def column_str(data_field, caption, **kwargs):
    return dict(dataField=data_field, caption=caption, dataType='string', **kwargs)


def column_int(data_field, caption, **kwargs):
    return dict(dataField=data_field, caption=caption, dataType='number', format='#0', **kwargs)


def column_float(data_field, caption, **kwargs):
    return dict(dataField=data_field, caption=caption, dataType='number',
                format=kwargs.pop('format', '#0.###'), **kwargs)


def column_dt(data_field, caption, **kwargs):
    return dict(dataField=data_field, caption=caption, dataType='datetime',
                format=kwargs.pop('format', 'dd.MM.yyyy HH:mm:ss'), **kwargs)


def column_d(data_field, caption, **kwargs):
    return dict(dataField=data_field, caption=caption, dataType='date',
                format=kwargs.pop('format', 'dd.MM.yyyy'), **kwargs)


ddx_table_columns = {
    forms.ModelChoiceField: column_str,
    # forms.BigIntegerField: column_int,  # IntegerField with min_value set to -9223372036854775808 and max_value set to 9223372036854775807.
    # forms.BinaryField: column_str,  # CharField, if editable is set to True on the model field, otherwise not represented in the form.
    forms.BooleanField: column_bool,  # BooleanField, or NullBooleanField if null=True.
    forms.CharField: column_str,  # CharField with max_length set to the model field’s max_length and empty_value set to None if null=True.
    forms.DateField: column_d,  # DateField
    forms.DateTimeField: column_dt,  # DateTimeField
    forms.DecimalField: column_float,  # DecimalField
    forms.DurationField: column_int,  # DurationField
    forms.EmailField: column_str,  # EmailField
    forms.FileField: column_str,  # FileField
    forms.FilePathField: column_str,  # FilePathField
    forms.FloatField: column_float,  # FloatField
    # forms.ForeignKey: column_str,  # ModelChoiceField (see below)
    forms.ImageField: column_str,  # ImageField
    forms.IntegerField: column_int,  # IntegerField
    # forms.IPAddressField: column_str,  # IPAddressField
    forms.GenericIPAddressField: column_str,  # GenericIPAddressField
    forms.JSONField: column_str,  # JSONField
    forms.ModelMultipleChoiceField: column_str,  # ModelMultipleChoiceField (see below)
    forms.NullBooleanField: column_bool,  # NullBooleanField
    # forms.PositiveBigIntegerField: column_int,  # IntegerField
    # forms.PositiveIntegerField: column_int,  # IntegerField
    # forms.PositiveSmallIntegerField: column_int,  # IntegerField
    forms.SlugField: column_str,  # SlugField
    # forms.SmallIntegerField: column_int,  # IntegerField
    # forms.TextField: column_str,  # CharField with widget=forms.Textarea
    forms.TimeField: column_float,  # TimeField
    forms.URLField: column_str,  # URLField
    forms.UUIDField: column_str,  # UUIDField
}


ddx_widgets = {
    forms.TextInput: ddx.TextBox,
    forms.URLInput: ddx.TextBox,
    forms.PasswordInput: ddx.TextBox,
    forms.HiddenInput: dcc.Input,
    forms.Textarea: ddx.TextArea,
    forms.DateInput: ddx.DateBox,
    forms.DateTimeInput: ddx.DateBox,
    forms.TimeInput: ddx.NumberBox,
    forms.CheckboxInput: ddx.CheckBox,
    forms.Select: ddx.SelectBox,
    forms.NumberInput: ddx.NumberBox,

    forms.ModelChoiceField: ddx.SelectBox,
    # forms.BigIntegerField: ddx.NumberBox,  # IntegerField with min_value set to -9223372036854775808 and max_value set to 9223372036854775807.
    # forms.BinaryField: ddx.TextBox,  # CharField, if editable is set to True on the model field, otherwise not represented in the form.
    forms.BooleanField: ddx.CheckBox,  # BooleanField, or NullBooleanField if null=True.
    forms.CharField: ddx.TextBox,  # CharField with max_length set to the model field’s max_length and empty_value set to None if null=True.
    forms.DateField: ddx.DateBox,  # DateField
    forms.DateTimeField: ddx.DateBox,  # DateTimeField
    forms.DecimalField: ddx.NumberBox,  # DecimalField
    forms.DurationField: ddx.NumberBox,  # DurationField
    forms.EmailField: ddx.TextBox,  # EmailField
    forms.FileField: ddx.TextBox,  # FileField
    forms.FilePathField: ddx.TextBox,  # FilePathField
    forms.FloatField: ddx.NumberBox,  # FloatField
    # forms.ForeignKey: ddx.SelectBox,  # ModelChoiceField (see below)
    forms.ImageField: ddx.TextBox,  # ImageField
    forms.IntegerField: ddx.NumberBox,  # IntegerField
    # forms.IPAddressField: ddx.TextBox,  # IPAddressField
    forms.GenericIPAddressField: ddx.TextBox,  # GenericIPAddressField
    forms.JSONField: ddx.TextArea,  # JSONField
    forms.ModelMultipleChoiceField: ddx.SelectBox,  # ModelMultipleChoiceField (see below)
    forms.NullBooleanField: ddx.CheckBox,  # NullBooleanField
    # forms.PositiveBigIntegerField: ddx.NumberBox,  # IntegerField
    # forms.PositiveIntegerField: ddx.NumberBox,  # IntegerField
    # forms.PositiveSmallIntegerField: ddx.NumberBox,  # IntegerField
    forms.SlugField: ddx.TextBox,  # SlugField
    # forms.SmallIntegerField: ddx.NumberBox,  # IntegerField
    # forms.TextField: ddx.TextBox,  # CharField with widget=forms.Textarea
    forms.TimeField: ddx.NumberBox,  # TimeField
    forms.URLField: ddx.TextBox,  # URLField
    forms.UUIDField: ddx.TextBox,  # UUIDField

    models.BigIntegerField: ddx.NumberBox,  # IntegerField with min_value set to -9223372036854775808 and max_value set to 9223372036854775807.
    models.BinaryField: ddx.TextBox,  # CharField, if editable is set to True on the model field, otherwise not represented in the form.
    models.BooleanField: ddx.CheckBox,  # BooleanField, or NullBooleanField if null=True.
    models.CharField: ddx.TextBox,  # CharField with max_length set to the model field’s max_length and empty_value set to None if null=True.
    models.DateField: ddx.DateBox,  # DateField
    models.DateTimeField: ddx.DateBox,  # DateTimeField
    models.DecimalField: ddx.NumberBox,  # DecimalField
    models.DurationField: ddx.NumberBox,  # DurationField
    models.EmailField: ddx.TextBox,  # EmailField
    models.FileField: ddx.TextBox,  # FileField
    models.FilePathField: ddx.TextBox,  # FilePathField
    models.FloatField: ddx.NumberBox,  # FloatField
    models.ForeignKey: ddx.SelectBox,  # ModelChoiceField (see below)
    models.ImageField: ddx.TextBox,  # ImageField
    models.IntegerField: ddx.NumberBox,  # IntegerField
    models.IPAddressField: ddx.TextBox,  # IPAddressField
    models.GenericIPAddressField: ddx.TextBox,  # GenericIPAddressField
    models.JSONField: ddx.TextArea,  # JSONField
    models.ManyToManyField: ddx.SelectBox,  # ModelMultipleChoiceField (see below)
    models.NullBooleanField: ddx.CheckBox,  # NullBooleanField
    models.PositiveBigIntegerField: ddx.NumberBox,  # IntegerField
    models.PositiveIntegerField: ddx.NumberBox,  # IntegerField
    models.PositiveSmallIntegerField: ddx.NumberBox,  # IntegerField
    models.SlugField: ddx.TextBox,  # SlugField
    models.SmallIntegerField: ddx.NumberBox,  # IntegerField
    models.TextField: ddx.TextArea,  # CharField with widget=forms.Textarea
    models.TimeField: ddx.NumberBox,  # TimeField
    models.URLField: ddx.TextBox,  # URLField
    models.UUIDField: ddx.TextBox,  # UUIDField
}


def display_for_field(value, field, empty_value_display):
    # from django.contrib.admin.templatetags.admin_list import _boolean_icon

    if getattr(field, 'flatchoices', None):
        return dict(field.flatchoices).get(value, empty_value_display)
    # BooleanField needs special-case null-handling, so it comes before the
    # general null test.
    elif isinstance(field, models.BooleanField):
        # return _boolean_icon(value)
        return value
    elif value is None:
        return empty_value_display
    elif isinstance(field, models.DateTimeField):
        # return formats.localize(timezone.template_localtime(value))
        return timezone.template_localtime(value)
    elif isinstance(field, (models.DateField, models.TimeField)):
        # return formats.localize(value)
        return value
    elif isinstance(field, models.DecimalField):
        # return formats.number_format(value, field.decimal_places)
        return value
    elif isinstance(field, (models.IntegerField, models.FloatField)):
        # return formats.number_format(value)
        return value
    # elif isinstance(field, models.FileField) and value:
    #     return format_html('<a href="{}">{}</a>', value.url, value)
    # elif isinstance(field, models.JSONField) and value:
    #     try:
    #         return json.dumps(value, ensure_ascii=False, cls=field.encoder)
    #     except TypeError:
    #         return display_for_value(value, empty_value_display)
    else:
        return display_for_value(value, empty_value_display)


def display_for_value(value, empty_value_display, boolean=False):
    # from django.contrib.admin.templatetags.admin_list import _boolean_icon

    if boolean:
        # return _boolean_icon(value)
        return value
    elif value is None:
        return empty_value_display
    elif isinstance(value, bool):
        # return str(value)
        return value
    elif isinstance(value, datetime.datetime):
        # return formats.localize(timezone.template_localtime(value))
        return timezone.template_localtime(value)
    elif isinstance(value, (datetime.date, datetime.time)):
        # return formats.localize(value)
        return value
    elif isinstance(value, (int, decimal.Decimal, float)):
        # return formats.number_format(value)
        return value
    elif isinstance(value, (list, tuple)):
        return ', '.join(str(v) for v in value)
    else:
        return str(value)



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='django-dash-ddx-catalog',
    version='0.1.0',
    description='Auxiliary Django forms for semi-automatic CRUD for dash-devextreme components',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Sergei Pikhovkin',
    author_email='s@pikhovkin.ru',
    url='https://github.com/pikhovkin/django-dash-ddx-catalog',
    packages=[
        'dash_ddx_catalog',
    ],
    include_package_data=True,
    install_requires=[
        'Django>=3.1,<4.0',
        'dj-plotly-dash[all]<0.18.0',
        'dash-devextreme',
        'dash_core_components<1.14.0',
        'dash_html_components<1.2.0',
    ],
    python_requires='>=3.7.*, <4.0.*',
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



import dash_html_components as html
import dash_devextreme as ddx

from django.test import TestCase
from django import forms

from dash_ddx_catalog.forms import DashDDXForm


class SimpleForm(DashDDXForm):
    name = forms.CharField(label='Test label', max_length=255)


class TestDashDDXForm(TestCase):
    def test_simple(self):
        fields = SimpleForm(initial=dict(name='Test', name2='Test2')).as_ddx_div()

        self.assertTrue(fields[0].__class__ is html.Div)
        self.assertTrue(fields[0].children[0].__class__ is html.Span)
        self.assertTrue(fields[0].children[1].__class__ is ddx.TextBox)
        self.assertTrue(fields[0].children[1].name == 'name')
        # self.assertTrue(fields[0].children[1].value == 'Test')
        # self.assertTrue(fields[0].children[1].maxLength == 255)



import dash_html_components as html

from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils import timezone
from django.contrib.admin.utils import lookup_field  #, display_for_value, display_for_field

from .fields import ddx_widgets, ddx_table_columns, display_for_value, display_for_field


__all__ = (
    'DashDDXForm',
    'DashDDXModelForm',
)


class DashDDXBaseForm(object):
    def as_ddx_div(self):
        tags = []
        # print('initial', self.initial)
        for f_name, f in self.base_fields.items():
            c = ddx_widgets[f.widget.__class__]

            label = str(f.label)
            is_required = f.required

            extra_attrs = f.widget_attrs(f.widget) or {}
            attrs = f.widget.attrs.copy()
            attrs.update(extra_attrs)

            ddx_attrs = {}
            if 'maxlength' in attrs:
                ddx_attrs['maxLength'] = attrs['maxlength']
            if 'min' in attrs:
                ddx_attrs['min'] = attrs['min']
            if 'max' in attrs:
                ddx_attrs['max'] = attrs['max']

            if c.__name__ == 'NumberBox':
                ddx_attrs['showSpinButtons'] = True
                ddx_attrs['showClearButton'] = True
                ddx_attrs['showSpinButtons'] = True
                ddx_attrs['step'] = 1
                if f.__class__.__name__ == 'FloatField':
                    ddx_attrs['format'] = '#0.##'
                else:
                    ddx_attrs['format'] = '#0'

            # print(f, f.__class__.__name__, f.label)
            # print(attrs)

            try:
                f, attr, value = lookup_field(f_name, self.instance)  # , cl.model_admin)
            except (ObjectDoesNotExist, AttributeError):
                value = None

            # print(f_name, value, type(value))

            if f.__class__.__name__ in ('DateTimeField', 'DateField'):
                if is_required:
                    v = (value or timezone.now()).isoformat()
                elif value:
                    v = value.isoformat()
                else:
                    v = None

                tags.append(
                    html.Div(className='option', children=[
                        html.Span(label),
                        c(name=f_name, type='datetime' if f.__class__.__name__ == 'DateTimeField' else 'date',
                          value=v,
                          displayFormat='dd.MM.yyyy HH:mm:ss', dateSerializationFormat='yyyy-MM-ddTHH:mm:ssZ',
                          openOnFieldClick=True, applyValueMode='useButtons', acceptCustomValue=True),
                    ])
                )
            elif f.__class__.__name__ in ('ForeignKey', 'ModelChoiceField'):
                related_objects = lambda obj: dict(id=obj.pk, name=str(obj))
                if hasattr(f, 'related_model'):
                    values = list(map(related_objects, f.related_model.objects.all()))
                else:
                    values = list(map(related_objects, f.queryset))

                if not is_required:
                    values.insert(0, dict(id=None, name=''))
                if is_required:
                    v = self.instance and value and value.pk or values and values[0]['id'] or None
                elif value:
                    v = value.pk
                else:
                    v = None

                tags.append(
                    html.Div(className='option', children=[
                        html.Span(label),
                        c(
                          dataSource={'store': values, 'key': 'id'}, name=f_name,
                          displayExpr='name', valueExpr='id', value=v,
                          searchEnabled=True, searchMode='contains', searchExpr='name',
                          searchTimeout=200,  # showDataBeforeSearchOption=True,
                        ),
                    ]),
                )
            elif f.__class__.__name__ == 'BooleanField':
                tags.append(
                    html.Div(className='option', children=[
                        html.Span(label, style=dict(marginRight=4)),
                        c(name=f_name, value=bool(value)),
                    ])
                )
            elif c.__name__ == 'TextArea':
                value = display_for_field(value, f, None)
                tags.append(
                    html.Div(className='option', children=[
                        html.Span(label),
                        c(name=f_name, value=value, height=48, **ddx_attrs),
                    ])
                )
            else:
                value = display_for_field(value, f, None)
                tags.append(
                    html.Div(className='option', children=[
                        html.Span(label),
                        c(name=f_name, value=value, **ddx_attrs),
                    ])
                )
        return tags

    @staticmethod
    def get_form_values(dash_form_children):
        data = {}
        while dash_form_children:
            elem = dash_form_children.pop(0)
            if 'children' in elem['props'] and elem['props']['children']:
                if isinstance(elem['props']['children'], list):
                    for c in elem['props']['children']:
                        dash_form_children.append(c)
                elif isinstance(elem['props']['children'], dict):
                    dash_form_children.append(elem['props']['children'])
            else:
                if elem['type'] in (
                        'Autocomplete', 'Calendar', 'CheckBox', 'ColorBox',
                        'DateBox', 'DropDownBox', 'HtmlEditor',
                        'Lookup', 'NumberBox', 'RadioGroup', 'RangeSlider',
                        'SelectBox', 'Slider', 'Switch',
                        'TagBox', 'TextArea', 'TextBox', 'Textarea', 'Input'
                ):
                    data[elem['props']['name']] = (
                        elem['props']['value']['value']
                        if isinstance(elem['props'].get('value'), dict)
                        else elem['props'].get('value')
                    )
        return data


class DashDDXForm(DashDDXBaseForm, forms.Form):
    ...


class DashDDXModelForm(DashDDXBaseForm, forms.ModelForm):
    @staticmethod
    def items_for_result(columns, result, form=None):
        for field_index, field_name in enumerate(columns):
            empty_value_display = None  #cl.model_admin.get_empty_value_display()
            # row_classes = ['field-%s' % _coerce_field_name(field_name, field_index)]
            try:
                f, attr, value = lookup_field(field_name, result)#, cl.model_admin)
            except ObjectDoesNotExist:
                result_repr = empty_value_display
            else:
                empty_value_display = getattr(attr, 'empty_value_display', empty_value_display)
                if f is None or f.auto_created:
                    # if field_name == 'action_checkbox':
                    #     row_classes = ['action-checkbox']
                    boolean = getattr(attr, 'boolean', False)
                    result_repr = display_for_value(value, empty_value_display, boolean)
                    # if isinstance(value, (datetime.date, datetime.time)):
                    #     row_classes.append('nowrap')
                else:
                    if isinstance(f.remote_field, models.ManyToOneRel):
                        field_val = getattr(result, f.name)
                        if field_val is None:
                            result_repr = empty_value_display
                        else:
                            result_repr = field_val
                    else:
                        result_repr = display_for_field(value, f, empty_value_display)
                    # if isinstance(f, (models.DateField, models.TimeField, models.ForeignKey)):
                    #     row_classes.append('nowrap')
            yield str(result_repr) if isinstance(result_repr, models.Model) else result_repr

    @classmethod
    def ddx_table_columns(cls):
        columns = []
        for f_name, f in cls.base_fields.items():
            Column = ddx_table_columns[f.__class__]
            columns.append(Column(f_name, str(f.label)))
        return columns

    @classmethod
    def ddx_table_rows(cls, *args, **kwargs):
        rows = []
        columns = ['id'] + [name for name in cls.base_fields.keys()]
        objects = cls._meta.model.objects.all().filter(*args, **kwargs)
        for obj in objects:
            values = cls.items_for_result(columns, obj)
            rows.append(dict(zip(columns, values)))
        return rows
